﻿using Lidgren.Network;
using Jeli;
using System;
using System.Threading;
using System.Collections.Generic;
using System.Diagnostics;

namespace Janus
{
	/// <summary>
	/// Triggered when timeline server is starting
	/// </summary>
	/// <param name="port">Port number on which timeline server listens</param>
	public delegate void ServerStartingHandler(int port);

	/// <summary>
	/// Networks to the timeline clients
	/// Establishes connections and processes incoming and outgoing messages
	/// </summary>
	public static class TimelineServer
	{
		static Dictionary<DeliveryMode, NetDeliveryMethod> _deliveryModes = new Dictionary<DeliveryMode, NetDeliveryMethod>()
		{
			{ DeliveryMode.ReliableOrdered, NetDeliveryMethod.ReliableOrdered},
			{ DeliveryMode.ReliableUnordered, NetDeliveryMethod.ReliableUnordered},
			{ DeliveryMode.Unreliable,NetDeliveryMethod.Unreliable}
		};

		/// <summary>
		/// Triggered when timeline server is starting
		/// </summary>
		public static event ServerStartingHandler ServerStarting;

		/// <summary>
		/// True if timeline server is running
		/// </summary>
		public static bool IsStarted
		{
			get { return _netServer != null && _netServer.Status == NetPeerStatus.Running; }
		}

		/// <summary>
		/// Provides a message interface to synchronizes multiple, remote timeline managers.
		/// </summary>
		public static TimelineSynchronizer TimelineSynchronizer
		{
			get { return _timelineSyncer; }
		}

		static NetServer _netServer;
		static TimelineSynchronizer _timelineSyncer;
		static bool _autoStep;
		static float _stepInterval;
		static Thread _workThread;
		static Dictionary<NetConnection, ushort> _peerIndices;
		static Dictionary<ushort, NetConnection> _peerConnections;

		static TimelineServer ()
		{
			_peerIndices = new Dictionary<NetConnection, ushort>();
			_peerConnections = new Dictionary<ushort, NetConnection>();
			_timelineSyncer = new TimelineSynchronizer();
		}

		/// <summary>
		/// Start timeline server - starts separate thread that processes incoming messages
		/// </summary>
		/// <param name="autoStep">If true, automatically increment time.  If false, time must be manually incremented by calling Step</param>
		public static bool Start(bool autoStep)
		{
			Config config = LoadSettings();

			_autoStep = autoStep;
			_stepInterval = 1f / config["StepRate"].IntValue;

			NetPeerConfiguration netConfig = new NetPeerConfiguration("janus");
			netConfig.MaximumConnections = config["MaxConnections"].IntValue;
			netConfig.Port = config["Port"].IntValue;
			netConfig.EnableMessageType(NetIncomingMessageType.ConnectionApproval);

			_netServer = new NetServer(netConfig);
			_netServer.Start();

			if (ServerStarting != null)
				ServerStarting(netConfig.Port);

			_workThread = new Thread(Work);
			_workThread.Start();

			return true;
		}

		static Config LoadSettings ()
		{
			Config config = Config.Load("TimelineServer.ini");

			if (config == null)
			{
				config = new Config();
				config.AddChild("Port", 14242);
				config.AddChild("MaxConnections", 32);
				config.AddChild("StepRate", 100);
				config.Save("TimelineServer.ini");
			}

			return config;
		}


		/// <summary>
		/// Processes incoming and outgoing messages
		/// </summary>
		public static void Update()
		{
			
			List<NetIncomingMessage> incomingMessages = new List<NetIncomingMessage>();
			_netServer.ReadMessages(incomingMessages);

			foreach (var message in incomingMessages)
			{
				if (message.MessageType == NetIncomingMessageType.Data)
				{
					try
					{
						TimelineMessage timelineMessage = new TimelineMessage(
							(TimelineMessageType)message.ReadByte(), message.ReadBytes(message.LengthBytes - 1));

						_timelineSyncer.ProcessIncomingMessage(_peerIndices[message.SenderConnection], timelineMessage);
					}
					catch
					{
						continue;
					}
				}
				else if (message.MessageType == NetIncomingMessageType.ConnectionApproval)
				{
					ushort peerIndex = GetNextFreePeerIndex();

					_peerIndices.Add(message.SenderConnection, peerIndex);
					_peerConnections.Add(peerIndex, message.SenderConnection);

					NetOutgoingMessage approval = _netServer.CreateMessage(sizeof(ushort));
					approval.Write(peerIndex);
					message.SenderConnection.Approve(approval);

					_timelineSyncer.ConnectPeer(peerIndex);
				}
				else if (message.MessageType == NetIncomingMessageType.StatusChanged)
				{
					NetConnectionStatus newStatus = (NetConnectionStatus)message.ReadByte();

					if (newStatus == NetConnectionStatus.Disconnected)
					{
						ushort peerIndex = _peerIndices[message.SenderConnection];

						_timelineSyncer.DisconnectPeer(peerIndex);

						_peerIndices.Remove(message.SenderConnection);
						_peerConnections.Remove(peerIndex);
					}
				}
			}
			
			foreach (var connPair in _peerConnections)
			{
				ushort peerIndex = connPair.Key;
				NetConnection netConnection = connPair.Value;

				// Don't send messages until we're connected.
				if (netConnection.Status != NetConnectionStatus.Connected)
					continue;

				foreach (var message in _timelineSyncer.GetOutgoingMessages(peerIndex))
				{
					if (message == null)
						continue;

					NetOutgoingMessage om = _netServer.CreateMessage(sizeof(byte) + message.Data.Length);
					om.Write((byte)message.MessageType);
					om.Write(message.Data);

					netConnection.SendMessage(om, _deliveryModes[message.DeliveryMode], 0);
				}
			}
		}

		/// <summary>
		/// Increments the current time
		/// </summary>
		/// <param name="deltaTime">Time step size in seconds</param>
		public static void Step(float deltaTime)
		{
			_timelineSyncer.Step(deltaTime);
		}

		static ushort GetNextFreePeerIndex ()
		{
			ushort peerIndex = 0;

			do
			{
				if (peerIndex == ushort.MaxValue)
					peerIndex = ushort.MinValue;
				else peerIndex++;
			}
			while (_peerConnections.ContainsKey(peerIndex));

			return peerIndex;
		}

		static void Work()
		{
			DateTime lastStepTime = DateTime.Now;

			while (_netServer.Status == NetPeerStatus.Running)
			{
				Update();

				if (_autoStep)
				{
					DateTime currentTime = DateTime.Now;
					float deltaTime = (float)currentTime.Subtract(lastStepTime).TotalSeconds;

					if (deltaTime >= _stepInterval)
					{
						Step(deltaTime);
						lastStepTime = currentTime;
					}
					Thread.Sleep(1);
				}
			}
		}

		/// <summary>
		/// Stops net server
		/// </summary>
		public static void Stop()
		{
			_netServer.Shutdown("");

			while (_netServer.Status != NetPeerStatus.NotRunning)
			{
				Thread.Sleep(1);
			}

			_netServer = null;
		}
	}
}